@@ -6,14 +6,13 @@ Created on Thu Sep 16 22:40:39 2021
 @author: daniel
 """
 import os
-import numpy as np
 from pathlib import Path
-import matplotlib.pyplot as plt
 from warnings import warn
+import matplotlib.pyplot as plt
+import numpy as np
 import pkg_resources
 import joblib
 
-os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
 from optuna.visualization.matplotlib import plot_optimization_history
 from tensorflow.keras.models import Sequential, save_model
 from tensorflow.keras.initializers import VarianceScaling
@@ -23,6 +22,7 @@ from tensorflow.keras.layers import Activation, Dense, Dropout, Conv2D, MaxPool2
 from tensorflow.keras.callbacks import ModelCheckpoint
 from keras.models import load_model
 
+os.environ['CUDA_VISIBLE_DEVICES'], os.environ['TF_CPP_MIN_LOG_LEVEL']= '-1', '3'
 from pyBIA.data_processing import process_class, create_training_set
 from pyBIA import optimization
 
@@ -70,8 +70,10 @@ class Classifier:
         Trained machine learning model.
 
     """
-    def __init__(self, blob_data=None, other_data=None, optimize=True, limit_search=True, metric='loss', n_iter=25, img_num_channels=1, normalize=True, min_pixel=638,
-        max_pixel=3000, val_X=None, val_Y=None, epochs=100, train_epochs=25, patience=5):
+    def __init__(self, blob_data=None, other_data=None, val_blob=None, val_other=None, img_num_channels=1, 
+        optimize=True, limit_search=True, n_iter=25, normalize=True, min_pixel=638, max_pixel=3000, epochs=100, train_epochs=25, 
+        patience=5, opt_aug=False, batch_min=10, batch_max=250, image_size_min=50, image_size_max=90, balance_val=True,
+        opt_min_pix=None, opt_max_pix=None, metric='loss'):
 
         self.blob_data = blob_data
         self.other_data = other_data
@@ -84,12 +86,21 @@ class Classifier:
         self.normalize = normalize 
         self.min_pixel = min_pixel
         self.max_pixel = max_pixel
-        self.val_X = val_X
-        self.val_Y = val_Y
+        self.val_blob = val_blob
+        self.val_other = val_other
         self.epochs = epochs
         self.train_epochs = train_epochs
         self.patience = patience
 
+        self.opt_aug = opt_aug
+        self.batch_min = batch_min 
+        self.batch_max = batch_max 
+        self.image_size_min = image_size_min
+        self.image_size_max = image_size_max
+        self.balance_val = balance_val
+        self.opt_min_pix = opt_min_pix
+        self.opt_max_pix = opt_max_pix
+
         self.model = None
         self.history = None 
         self.best_params = None 
@@ -97,43 +108,57 @@ class Classifier:
 
     def create(self):
         """
-        Creates the machine learning engine, current options are either a
-        Random Forest, XGBoost, or a Neural Network classifier. 
+        Creates the CNN machine learning engine.
         
         Returns:
-            Trained and optimized classifier.
+            Trained classifier.
         """
         
         if self.optimize is False:
             print("Returning base model...")
             self.model, self.history = pyBIA_model(self.blob_data, self.other_data, img_num_channels=self.img_num_channels, normalize=self.normalize,
-                min_pixel=self.min_pixel, max_pixel=self.max_pixel, val_X=self.val_X, val_Y=self.val_Y, epochs=self.epochs)
+                min_pixel=self.min_pixel, max_pixel=self.max_pixel, val_blob=self.val_blob, val_other=self.val_other, epochs=self.epochs)
             return      
 
         self.best_params, self.optimization_results = optimization.hyper_opt(self.blob_data, self.other_data, clf='cnn', metric=self.metric, n_iter=self.n_iter, 
             balance=False, return_study=True, img_num_channels=self.img_num_channels, normalize=self.normalize, min_pixel=self.min_pixel, 
-            max_pixel=self.max_pixel, val_X=self.val_X, val_Y=self.val_Y, train_epochs=self.train_epochs, patience=self.patience, limit_search=self.limit_search)
+            max_pixel=self.max_pixel, val_X=self.val_blob, val_Y=self.val_other, train_epochs=self.train_epochs, patience=self.patience, limit_search=self.limit_search,
+            opt_aug=self.opt_aug, batch_min=self.batch_min, batch_max=self.batch_max, image_size_min=self.image_size_min, image_size_max=self.image_size_max, balance_val=self.balance_val,
+            opt_min_pix=self.opt_min_pix, opt_max_pix=self.opt_max_pix)
 
         print("Fitting and returning final model...")
-        if limit_search:
+        if self.opt_min_pix is not None:
+            min_pix, max_pix = self.best_params['min_pix'], self.best_params['max_pix']
+        else:
+            min_pix, max_pix = self.min_pixel, self.max_pixel
+
+        if self.limit_search:
             self.model, self.history = pyBIA_model(self.blob_data, self.other_data, img_num_channels=self.img_num_channels, normalize=self.normalize,
-                min_pixel=self.min_pixel, max_pixel=self.max_pixel, val_X=self.val_X, val_Y=self.val_Y, epochs=self.epochs, batch_size=self.best_params['batch_size'],
-                lr=self.best_params['lr'], loss=self.best_params['loss'])
+                min_pixel=min_pix, max_pixel=max_pix, val_blob=self.val_blob, val_other=self.val_other, epochs=self.epochs, batch_size=self.best_params['batch_size'],
+                lr=self.best_params['lr'], decay=self.best_params['decay'], maxpool_size_1=self.best_params['maxpool_size_1'], maxpool_stride_1=self.best_params['maxpool_stride_1'],
+                maxpool_size_2=self.best_params['maxpool_size_2'], maxpool_stride_2=self.best_params['maxpool_stride_2'], maxpool_size_3=self.best_params['maxpool_size_3'], 
+                maxpool_stride_3=self.best_params['maxpool_stride_3'], filter_1=self.best_params['filter_1'], filter_size_1=self.best_params['filter_size_1'], strides_1=self.best_params['strides_1'],
+                filter_2=self.best_params['filter_2'], filter_size_2=self.best_params['filter_size_2'], strides_2=self.best_params['strides_2'], filter_3=self.best_params['filter_3'], 
+                filter_size_3=self.best_params['filter_size_3'],  strides_3=self.best_params['strides_3'], filter_4=self.best_params['filter_4'], filter_size_4=self.best_params['filter_size_4'], 
+                strides_4=self.best_params['strides_4'], filter_5=self.best_params['filter_5'], filter_size_5=self.best_params['filter_size_5'],  strides_5=self.best_params['strides_5'])
         else:
             self.model, self.history = pyBIA_model(self.blob_data, self.other_data, img_num_channels=self.img_num_channels, normalize=self.normalize,
-                min_pixel=self.min_pixel, max_pixel=self.max_pixel, val_X=self.val_X, val_Y=self.val_Y, epochs=self.epochs, batch_size=self.best_params['batch_size'],
-                lr=self.best_params['lr'], batch_norm=self.best_params['batch_norm'], momentum=self.best_params['momentum'],
-                decay=self.best_params['decay'], nesterov=self.best_params['nesterov'], loss=self.best_params['loss'],
-                padding=self.best_params['padding'], dropout_1=self.best_params['dropout_1'], dropout_2=self.best_params['dropout_2'],
-                activation_conv=self.best_params['activation_conv'], activation_dense=self.best_params['activation_dense'],
-                maxpool_size=self.best_params['maxpool_size'], maxpool_stride=self.best_params['maxpool_stride']) #params
+                min_pixel=min_pix, max_pixel=max_pix, val_blob=self.val_blob, val_other=self.val_other, epochs=self.epochs, batch_size=self.best_params['batch_size'],
+                lr=self.best_params['lr'], momentum=self.best_params['momentum'], decay=self.best_params['decay'], nesterov=self.best_params['nesterov'], 
+                dropout=self.best_params['dropout'], activation_conv=self.best_params['activation_conv'], activation_dense=self.best_params['activation_dense'], 
+                maxpool_size_1=self.best_params['maxpool_size_1'], maxpool_stride_1=self.best_params['maxpool_stride_1'], maxpool_size_2=self.best_params['maxpool_size_2'], 
+                maxpool_stride_2=self.best_params['maxpool_stride_2'], maxpool_size_3=self.best_params['maxpool_size_3'], maxpool_stride_3=self.best_params['maxpool_stride_3'],
+                filter_1=self.best_params['filter_1'], filter_size_1=self.best_params['filter_size_1'], strides_1=self.best_params['strides_1'], filter_2=self.best_params['filter_2'], 
+                filter_size_2=self.best_params['filter_size_2'], strides_2=self.best_params['strides_2'], filter_3=self.best_params['filter_3'], filter_size_3=self.best_params['filter_size_3'], 
+                strides_3=self.best_params['strides_3'], filter_4=self.best_params['filter_4'], filter_size_4=self.best_params['filter_size_4'], strides_4=self.best_params['strides_4'], 
+                filter_5=self.best_params['filter_5'], filter_size_5=self.best_params['filter_size_5'],  strides_5=self.best_params['strides_5'])
             
         return 
 
     def save(self, path=None, overwrite=False):
         """
         Saves the trained classifier in a new directory named 'pyBIA_models', 
-        as well as the imputer and the features to use attributes, if not None.
+        as well as the imputer and the features to use, if applicable.
         
         Args:
             path (str): Absolute path where the data folder will be saved
@@ -143,6 +168,7 @@ class Classifier:
                 function creates in the specified path will be deleted if it exists
                 and created anew to avoid duplicate files. 
         """
+
         if self.model is None:
             raise ValueError('The models have not been created! Run classifier.create() first.')
 
@@ -169,7 +195,7 @@ class Classifier:
         if self.model is not None:      
             np.savetxt(path+'model_acc', self.history.history['accuracy'])
             np.savetxt(path+'model_loss', self.history.history['loss'])
-            if self.val_X is not None:
+            if self.val_blob is not None:
                 np.savetxt(path+'model_val_acc', self.history.history['val_accuracy'])
                 np.savetxt(path+'model_val_loss', self.history.history['val_loss'])
 
@@ -244,7 +270,6 @@ class Classifier:
                 which there is an 'OTHER' class. Defaults to 'DIFFUSE'. 
         Returns:
             The class prediction(s), either 'DIFFUSE' or 'OTHER'.
-
         """
       
         data = process_class(data, normalize=self.normalize, min_pixel=self.min_pixel, max_pixel=self.max_pixel, img_num_channels=self.img_num_channels)
@@ -261,7 +286,7 @@ class Classifier:
 
         return np.array(output)
 
-    def plot_hyper_opt(self, xlog=True, ylog=False):
+    def plot_hyper_opt(self, xlog=True, ylog=False, savefig=False):
         """
         Plots the hyperparameter optimization history.
     
@@ -280,8 +305,20 @@ class Classifier:
             plt.xscale('log')
         if ylog:
             plt.yscale('log')
+
+        if self.metric == 'val_accuracy':
+            ylabel = 'Validation Accuracy'
+        elif self.metric == 'accuracy':
+            ylabel = 'Training Accuracy'
+        elif self.metric == 'val_loss':
+            ylabel = 'Validation Loss'
+        elif self.metric == 'loss':
+            ylabel = 'Training Loss'
+        else:
+            ylabel = 'Optimization Metric'
+
         plt.xlabel('Trial #', size=16)
-        plt.ylabel('10-fold CV Accuracy', size=16)
+        plt.ylabel(ylabel, size=16)
         plt.title(('Hyperparameter Optimization History'), size=18)
         #plt.xlim((1,1e4))
         #plt.ylim((0.9, 0.935))
@@ -289,43 +326,47 @@ class Classifier:
         plt.yticks(fontsize=14)
         plt.grid(True, color='k', alpha=0.35, linewidth=1.5, linestyle='--')
         plt.legend(prop={'size': 16})
-        plt.show()
+        if savefig:
+            plt.savefig('CNN_Hyperparameter_Optimization.png', bbox_inches='tight', dpi=300)
+        else:
+            plt.show()
 
-    def load_bw_model(self):
-        """
-        Calling this will load the trained Tensorflow model, trained using NDWFS images
-        in the blue broadband.
-        
-        Note:
-            Training new models with 1000 epochs can take over a week, this Bw model
-            was trained using NDWFS blue broadband images of the Bootes field. The 
-            corresponding .h5 file is located in the data folder inside the pyBIA directory 
-            in the Python path. 
+    #def load_bw_model(self):
+    #    """
+    #    Calling this will load the trained Tensorflow model, trained using NDWFS images
+    #    in the blue broadband.
+    #    
+    #    Note:
+    #        Training new models with 1000 epochs can take over a week, this Bw model
+    #        was trained using NDWFS blue broadband images of the Bootes field. The 
+    #        corresponding .h5 file is located in the data folder inside the pyBIA directory 
+    #       in the Python path. 
+    #
+    #    Returns:
+    #        The pyBIA CNN model used for classifying images in blue broadband surveys.
 
-        Returns:
-            The pyBIA CNN model used for classifying images in blue broadband surveys.
+    #  """
 
-        """
-
-
-        resource_package = __name__
-        resource_path = '/'.join(('data', 'Bw_CNN_Model.h5'))
-        self.model = load_model(pkg_resources.resource_filename(resource_package, resource_path))
-        print('Bw model successfully loaded.')
-        print('Note: Input data when using this model must be 50x50.')
-        return 
+    #   resource_package = __name__
+    #   resource_path = '/'.join(('data', 'Bw_CNN_Model.h5'))
+    #   self.model = load_model(pkg_resources.resource_filename(resource_package, resource_path))
+    #   print('Bw model successfully loaded.')
+    #   print('Note: Input data when using this model must be 50x50.')
+    #   return 
 
 def pyBIA_model(blob_data, other_data, img_num_channels=1, normalize=True, 
-        min_pixel=0, max_pixel=3000, val_X=None, val_Y=None, epochs=100, 
-        batch_size=32, lr=0.001, batch_norm=True, momentum=0.9, decay=0.000083, 
-        nesterov=False, loss='categorical_crossentropy', activation_conv='relu', 
-        activation_dense='tanh', padding='same', dropout_1=0.5, dropout_2=0.5, 
-        pooling=True, maxpool_size=3, maxpool_stride=2, early_stop_callback=None, 
-        checkpoint=True):
+        min_pixel=0, max_pixel=100, val_blob=None, val_other=None, epochs=100, 
+        batch_size=32, lr=0.0001, momentum=0.9, decay=0.0, nesterov=False, 
+        loss='categorical_crossentropy', activation_conv='relu', activation_dense='relu', 
+        padding='same', dropout=0.5, pooling=True, maxpool_size_1=3, maxpool_stride_1=2,
+        maxpool_size_2=3, maxpool_stride_2=2, maxpool_size_3=3, maxpool_stride_3=2, 
+        filter_1=96, filter_size_1=11, strides_1=4, filter_2=256, filter_size_2=5, strides_2=1,
+        filter_3=384, filter_size_3=3, strides_3=1, filter_4=384, filter_size_4=3, strides_4=1,
+        filter_5=256, filter_size_5=3, strides_5=1, early_stop_callback=None, checkpoint=True):
         """
         The CNN model infrastructure presented by the 2012 ImageNet Large Scale 
         Visual Recognition Challenge, AlexNet. Parameters were adapted for
-        our astronomy case of detecting diffuse objects.
+        our astronomy case of detecting diffuse emission.
 
         Note:
             To avoid exploding gradients we need to normalize our pixels to be 
@@ -351,18 +392,14 @@ def pyBIA_model(blob_data, other_data, img_num_channels=1, normalize=True,
                 Pixels with counts below this threshold will be set to this limit.
             max_pixel (int, optional): The maximum pixel count, defaults to 3000. 
                 Pixels with counts above this threshold will be set to this limit.
-            val_X (array, optional): 3D matrix containing the 2D arrays (images)
-                to be used for validation.
-            val_Y (array, optional): A binary class matrix containing the labels of the
-                corresponding validation data. This binary matrix representation can be created
-                using tensorflow, see example in the Notes.
+            val_blob (array, optional): 3D matrix containing the 2D arrays (images)
+                to be used for validationm, for the blob class. Defaults to None.
+            val_other (array, optional): 3D matrix containing the 2D arrays (images)
+                to be used for validationm, for the blob class. Defaults to None.
             epochs (int): Number of epochs used for training. 
             batch_size (int): The size of each sub-sample used during the training
                 epoch. Large batches are likely to get stuck in local minima. Defaults to 32.
             lr (float): Learning rate, the rate at which the model updates the gradient. Defaults to 0.0001
-            batch_norm (bool): If False, batch normalization after each max pooling layer
-                is disabled. Defaults to True.
-                Note: If set to True it might be best also set dropout=0. See: https://arxiv.org/abs/1502.03167
             momentum (float): Momentum is a float greater than 0 that accelerates gradient descent. Defaults to 0.9.
             decay (float): The rate of learning rate decay applied after each epoch. Defaults to 0.0005. It is recommended
                 to set decay to the learning rate divded by the total number of epochs.
@@ -374,12 +411,8 @@ def pyBIA_model(blob_data, other_data, img_num_channels=1, normalize=True,
             padding (str): Either 'same' or 'valid'. When set to 'valid', the dimensions reduce as the boundary 
                 that doesn't make it within even convolutions get cuts off. Defaults to 'same', which applies
                 zero-value padding around the boundary, ensuring even convolutional steps across each dimension.
-            dropout_1 (float): Droupout rate after the first dense layer. This is the percentage of dense neurons
+            dropout (float): Droupout rate after the dense layers. This is the percentage of dense neurons
                 that are turned off at each epoch. This prevents inter-neuron depedency, and thus overfitting. 
-                Note: If batch_norm=True, it might be best to disable droput. See: https://arxiv.org/abs/1502.03167
-            dropout_2 (float): Droupout rate after the second dense layer. This is the percentage of dense neurons
-                that are turned off at each epoch. This prevents inter-neuron depedency, and thus overfitting. 
-                Note: If batch_norm=True, it might be best to disable droput. See: https://arxiv.org/abs/1502.03167
             pooling (bool): True to enable max pooling, false to disable. 
                 Note: Max pooling can result in loss of positional information, it computation allows
                 setting pooling=False may yield more robust accuracy.
@@ -388,34 +421,6 @@ def pyBIA_model(blob_data, other_data, img_num_channels=1, normalize=True,
             early_stop_callback (list, optional): Callbacks for early stopping and pruning with Optuna, defaults
                 to None. Should only be used during optimization, refer to pyBIA.optimization.objective_cnn().
             checkpoint (bool, optional): If False no checkpoint will be saved. Defaults to True.
-        Example:
-            To use a validation dataset when training the model, the val_X and val_Y
-            parameters must be input. The val_X is a 3D matrix containing all the images, and
-            the val_Y is another matrix containing their class label (0 for DIFFUSE, 1 for OTHER).
-
-            If you have validation data of blobs and others, we can use the pyBIA data_processing module to
-            properly process our data and costruct validation arguments of appropriate shape. 
-
-                >>> blob_x, blob_y = pyBIA.data_processing.process_class(blob_val, label=0, normalize=False)
-                >>> other_x, other_x = pyBIA.data_processing.process_class(other_val, label=1, normalize=False)
-        
-                >>> val_X = np.r_[blob_x, other_x]
-                >>> val_Y = np.r_[blob_y, other_y]
-
-                >>> model = create(blob_train, other_train, val_X=val_X, val_Y=val_Y)
-
-            The process_class function will reshape our data and label array, as per the CNN requirements.
-
-            If need-be, this reshaped data array can be constructed manually as follows:
-
-                >>> data = channel.reshape(axis, img_width, img_height, img_num_channels)
-
-            The label array is also a special binary form: you can set this up manually by converting your 1D label arrays 
-            into a binary matrix representation. For example, if you have a channel with 100 DIFFUSE samples (label=0), 
-            you can create a corresponding label array using numpy and keras:
-
-                >>> label_array = numpy.expand_dims(np.array([0]*len(channel)), axis=1)
-                >>> label_array = tensorflow.keras.utils.to_categorical(label, 2)
             
         Returns:
             The trained CNN model.
@@ -423,18 +428,22 @@ def pyBIA_model(blob_data, other_data, img_num_channels=1, normalize=True,
         
         if len(blob_data.shape) != len(other_data.shape):
             raise ValueError("Shape of blob and other data must be the same.")
-        if val_X is not None:
-            if val_Y is None:
-                raise ValueError("Need to input validation data labels (val_Y).")
-        if val_Y is not None:
-            if val_X is None:
-                raise ValueError("Need to input validation data (val_X).")
-        if val_X is not None:
-            if len(val_X) != len(val_Y):
-                raise ValueError("Size of validation data and validation labels must be the same.")
         if batch_size < 16:
-            warn("Batch Normalization can be unstable with low batch sizes, if loss returns nan try a larger batch size and/or smaller learning rate, otherwise set batch_norm=False.", stacklevel=2)
-        
+            warn("Batch Normalization can be unstable with low batch sizes, if loss returns nan try a larger batch size and/or smaller learning rate.", stacklevel=2)
+        if val_blob is not None:
+            val_X1, val_Y1 = process_class(val_blob, label=0, min_pixel=min_pixel, max_pixel=max_pixel, img_num_channels=img_num_channels)
+            if val_other is None:
+                val_X, val_Y = val_X1, val_Y1
+            else:
+                val_X2, val_Y2 = process_class(val_other, label=1, min_pixel=min_pixel, max_pixel=max_pixel, img_num_channels=img_num_channels)
+                val_X, val_Y = np.r_[val_X1, val_X2], np.r_[val_Y1, val_Y2]
+        else:
+            if val_other is None:
+                val_X, val_Y = None, None
+            else:
+                val_X2, val_Y2 = process_class(val_other, label=1, min_pixel=min_pixel, max_pixel=max_pixel, img_num_channels=img_num_channels)
+                val_X, val_Y = val_X2, val_Y2
+
         img_width = blob_data[0].shape[0]
         img_height = blob_data[0].shape[1]
         #decay = lr / epochs
@@ -447,8 +456,10 @@ def pyBIA_model(blob_data, other_data, img_num_channels=1, normalize=True,
 
         X_train, Y_train = create_training_set(blob_data, other_data, normalize=normalize, min_pixel=min_pixel, max_pixel=max_pixel, img_num_channels=img_num_channels)
 
-        X_train[X_train > 1] = 1
-        X_train[X_train < 0] = 0
+        if normalize:
+            X_train[X_train > 1] = 1
+            X_train[X_train < 0] = 0
+            
         input_shape = (img_width, img_height, img_num_channels)
        
         # Uniform scaling initializer
@@ -462,38 +473,41 @@ def pyBIA_model(blob_data, other_data, img_num_channels=1, normalize=True,
         # Model configuration
         model = Sequential()
 
-        model.add(Conv2D(96, 11, strides=4, activation=activation_conv, input_shape=input_shape,
+        #Convolutional layers
+        model.add(Conv2D(filter_1, filter_size_1, strides=strides_1, activation=activation_conv, input_shape=input_shape,
                          padding=padding, kernel_initializer=uniform_scaling))
         if pooling:
-            model.add(MaxPool2D(pool_size=maxpool_size, strides=maxpool_stride, padding=padding))
-        if batch_norm:
-            model.add(BatchNormalization())
+            model.add(MaxPool2D(pool_size=maxpool_size_1, strides=maxpool_stride_1, padding=padding))
+        model.add(BatchNormalization())
 
-        model.add(Conv2D(256, 5, activation=activation_conv, padding=padding,
+        model.add(Conv2D(filter_2, filter_size_2, strides=strides_2, activation=activation_conv, padding=padding,
                          kernel_initializer=uniform_scaling))
         if pooling:
-            model.add(MaxPool2D(pool_size=maxpool_size, strides=maxpool_stride, padding=padding))
-        if batch_norm:
-            model.add(BatchNormalization())
+            model.add(MaxPool2D(pool_size=maxpool_size_2, strides=maxpool_stride_2, padding=padding))
+        model.add(BatchNormalization())
 
-        model.add(Conv2D(384, 3, activation=activation_conv, padding=padding,
+        model.add(Conv2D(filter_3, filter_size_3, strides=strides_3, activation=activation_conv, padding=padding,
                          kernel_initializer=uniform_scaling))
-        model.add(Conv2D(384, 3, activation=activation_conv, padding=padding,
+        model.add(BatchNormalization())
+        model.add(Conv2D(filter_4, filter_size_4, strides=strides_4, activation=activation_conv, padding=padding,
                          kernel_initializer=uniform_scaling))
-        model.add(Conv2D(256, 3, activation=activation_conv, padding=padding,
+        model.add(BatchNormalization())
+        model.add(Conv2D(filter_5, filter_size_5, strides=strides_5, activation=activation_conv, padding=padding,
                          kernel_initializer=uniform_scaling))
         if pooling:
-            model.add(MaxPool2D(pool_size=maxpool_size, strides=maxpool_stride, padding=padding))
-        if batch_norm:
-            model.add(BatchNormalization())
+            model.add(MaxPool2D(pool_size=maxpool_size_3, strides=maxpool_stride_3, padding=padding))
+        model.add(BatchNormalization())
 
+        #FCC
         model.add(Flatten())
         model.add(Dense(4096, activation=activation_dense,
                         kernel_initializer='TruncatedNormal'))
-        model.add(Dropout(dropout_1))
+        model.add(Dropout(dropout))
         model.add(Dense(4096, activation=activation_dense,
                         kernel_initializer='TruncatedNormal'))
-        model.add(Dropout(dropout_2))
+        model.add(Dropout(dropout))
+
+        #Output layer
         model.add(Dense(num_classes, activation='softmax',
                         kernel_initializer='TruncatedNormal'))
 
@@ -509,9 +523,9 @@ def pyBIA_model(blob_data, other_data, img_num_channels=1, normalize=True,
             callbacks_list.append(early_stop_callback)
 
         if val_X is None:
-            history = model.fit(X_train, Y_train, batch_size=batch_size, epochs=epochs, callbacks=callbacks_list, verbose=2)
+            history = model.fit(X_train, Y_train, batch_size=batch_size, epochs=epochs, callbacks=callbacks_list, verbose=1)
         else:
-            history = model.fit(X_train, Y_train, batch_size=batch_size, validation_data=(val_X, val_Y), epochs=epochs, callbacks=callbacks_list, verbose=2)
+            history = model.fit(X_train, Y_train, batch_size=batch_size, validation_data=(val_X, val_Y), epochs=epochs, callbacks=callbacks_list, verbose=1)
 
         return model, history
 
