@@ -21,10 +21,16 @@ def augmentation(channel1, channel2=None, channel3=None, batch=10, width_shift=5
     This function takes in one image and applies data augmentation techniques.
     Shifts and rotations occur at random, for example, if width_shift is set
     to 10, then an image shift between -10 and 10 pixels will be chosen from a 
-    random uniform distribution.
+    random uniform distribution. Rotation angle is also chosen from a random uniform 
+    distribution, between zero and the rotation argument. 
 
-    Rotation angle is also chosen from a random uniform distribution, between
-    zero and the rotation argument. 
+    Note:
+        This function is used for offline data augmentation! In practice,
+        online augmentation may be preferred as that exposes the CNN
+        to significantly more samples. If multiple channels are input,
+        this method will save the seeds from the augmentation of the first 
+        channel, after which the seeds will be applied to the remaining channels,
+        thus ensuring the same augmentation procedure is applied across all filters.
 
     Args:
         channel1 (ndarray): 2D array of containing a single image, or a 3D array containing
@@ -150,7 +156,7 @@ def augmentation(channel1, channel2=None, channel3=None, batch=10, width_shift=5
 def resize(data, size=50):
     """
     Resizes the data by cropping out the outer boundaries outside the size x size limit.
-    Can be either 2D for one sample, or 3D for multiple samples.
+    Can be either a 2D array containing one sample, or a 3D array for multiple samples.
 
     Args:
         data (array): 2D array
@@ -203,7 +209,7 @@ def plot(data, cmap='gray', title=''):
     Args:
         data (array): 2D array for single image, or 3D array with stacked channels.
         cmap (str): Colormap to use when generating the image.
-            title (str, optional): Title displayed above the image. 
+        title (str, optional): Title displayed above the image. 
 
     Returns:
         AxesImage.
