@@ -25,6 +25,7 @@ from sklearn.model_selection import KFold, StratifiedKFold, train_test_split
 from sklearn.manifold import TSNE
 
 from pyBIA.optimization import hyper_opt, borutashap_opt, KNN_imputation, MissForest_imputation
+from optuna.visualization.matplotlib import plot_optimization_history
 from xgboost import XGBClassifier
 import scikitplot as skplt
 
@@ -77,12 +78,14 @@ class Classifier:
         Trained machine learning model.
 
     """
-    def __init__(self, data_x=None, data_y=None, clf='rf', optimize=True, impute=True, imp_method='KNN', 
+    def __init__(self, data_x=None, data_y=None, clf='rf', optimize=True, limit_search=True, impute=True, imp_method='KNN', 
         n_iter=25, boruta_trials=50, balance=True):
+
         self.data_x = data_x
         self.data_y = data_y
         self.clf = clf
         self.optimize = optimize 
+        self.limit_search = limit_search
         self.impute = impute
         self.imp_method = imp_method
         self.n_iter = n_iter
@@ -95,6 +98,22 @@ class Classifier:
 
         self.feature_history = None  
         self.optimization_results = None 
+        self.best_params = None 
+
+        if self.data_x is None or self.data_y is None:
+            print('NOTE: data_x and data_y parameters are required if you wish to output visualizations.')
+        
+        if self.clf == 'xgb':
+            if all(isinstance(val, (int, str)) for val in self.data_y):
+                print('XGBoost classifier requires numerical class labels! Converting class labels as follows:')
+                print('________________________________')
+                y = np.zeros(len(self.data_y))
+                for i in range(len(np.unique(self.data_y))):
+                    print(str(np.unique(self.data_y)[i]).ljust(10)+'  ------------->     '+str(i))
+                    index = np.where(self.data_y == np.unique(self.data_y)[i])[0]
+                    y[index] = i
+                self.data_y = y 
+                print('________________________________')
 
     def create(self):
         """
@@ -122,11 +141,11 @@ class Classifier:
                 print('________________________________')
                 y = np.zeros(len(self.data_y))
                 for i in range(len(np.unique(self.data_y))):
-                    print(str(np.unique(data_y)[i]).ljust(10)+'  ------------->     '+str(i))
+                    print(str(np.unique(self.data_y)[i]).ljust(10)+'  ------------->     '+str(i))
                     index = np.where(self.data_y == np.unique(self.data_y)[i])[0]
                     y[index] = i
                 self.data_y = y 
-                print('--------------------------------')
+                print('________________________________')
 
         else:
             raise ValueError('clf argument must either be "rf", "nn", or "xgb".')
@@ -166,8 +185,8 @@ class Classifier:
         else: 
             self.data_x = self.data_x[:,self.feats_to_use]
 
-        self.model, best_params, self.optimization_results = hyper_opt(self.data_x, self.data_y, clf=self.clf, n_iter=self.n_iter, 
-            balance=self.balance, return_study=True)
+        self.model, self.best_params, self.optimization_results = hyper_opt(self.data_x, self.data_y, clf=self.clf, n_iter=self.n_iter, 
+            balance=self.balance, return_study=True, limit_search=self.limit_search)
         print("Fitting and returning final model...")
         self.model.fit(self.data_x, self.data_y)
         
@@ -197,7 +216,7 @@ class Classifier:
         except FileExistsError:
             if overwrite:
                 try:
-                    os.rmdir(path+'pyBIA_models')
+                    os.rmdir(path+'pyBIA_ensemble_model')
                 except OSError:
                     for file in os.listdir(path+'pyBIA_ensemble_model'):
                         os.remove(path+'pyBIA_ensemble_model/'+file)
@@ -215,6 +234,8 @@ class Classifier:
             joblib.dump(self.feats_to_use, path+'Feats_Index')
         if self.optimization_results is not None:
             joblib.dump(self.optimization_results, path+'HyperOpt_Results')
+        if self.best_params is not None:
+            joblib.dump(self.best_params, path+'Best_Params')
         print('Files saved in: {}'.format(path))
         return 
 
@@ -235,7 +256,7 @@ class Classifier:
         if path[-1] != '/':
             path+='/'
 
-        path += 'pyBIA_models/'
+        path += 'pyBIA_ensemble_model/'
 
         try:
             self.model = joblib.load(path+'Model')
@@ -259,7 +280,7 @@ class Classifier:
             pass
 
         try:
-            self.optimization_results = joblib.load(path+'Optimization_Results')
+            self.optimization_results = joblib.load(path+'HyperOpt_Results')
             optimization_results = 'optimization_results'
         except FileNotFoundError:
             optimization_results = '' 
@@ -321,7 +342,8 @@ class Classifier:
             
         return np.array(output)
 
-    def plot_tsne(self, norm=True, pca=False, title='Feature Parameter Space'):
+    def plot_tsne(self, norm=True, pca=False, title='Feature Parameter Space', 
+        savefig=False):
         """
         Plots a t-SNE projection using the sklearn.manifold.TSNE() method.
 
@@ -371,13 +393,18 @@ class Classifier:
             mask = np.where(self.data_y == feat)[0]
             plt.scatter(x[mask], y[mask], marker=markers[count], label=str(feat), alpha=0.7)
 
-        plt.legend(loc='upper right', prop={'size': 16})
+        plt.legend(prop={'size': 16})
         plt.title(title, size=18)
         plt.xticks(fontsize=14)
         plt.yticks(fontsize=14)
-        plt.show()
+        if savefig:
+            plt.savefig('tSNE_Projection.png', bbox_inches='tight', dpi=300)
+            plt.clf()
+        else:
+            plt.show()
 
-    def plot_conf_matrix(self, norm=True, pca=False, k_fold=10, normalize=True, title='Confusion matrix'):
+    def plot_conf_matrix(self, norm=True, pca=False, k_fold=10, normalize=True, 
+        title='Confusion Matrix', savefig=False):
         """
         Returns a confusion matrix with k-fold validation.
 
@@ -404,6 +431,7 @@ class Classifier:
         Returns:
             AxesImage.
         """
+        
         classes = [str(label) for label in np.unique(self.data_y)]
         if np.any(np.isnan(self.data_x)):
             print('Automatically imputing NaN values with KNN imputation.')
@@ -422,21 +450,14 @@ class Classifier:
             data = np.asarray(pca_data).astype('float64')
 
         predicted_target, actual_target = evaluate_model(self.model, data, self.data_y, normalize=normalize, k_fold=k_fold)
-        generate_matrix(predicted_target, actual_target, normalize=normalize, classes=classes, title=title)
+        generate_matrix(predicted_target, actual_target, classes, normalize=normalize, title=title, savefig=savefig)
 
-    def plot_roc_curve(self, k_fold=10, title="Receiver Operating Characteristic Curve"):
+    def plot_roc_curve(self, k_fold=10, title="Receiver Operating Characteristic Curve", 
+        savefig=False):
         """
         Plots ROC curve with k-fold cross-validation, as such the 
         standard deviation variations are plotted.
         
-        Example:
-            To assess the performance of a random forest classifier (created
-            using the scikit-learn implementation) we can run the following:
-            
-            >>> from sklearn.ensemble import RandomForestClassifier
-            >>> classifier = RandomForestClassifier()
-            >>> plot_roc_curve(classifier, data_x, data_y)
-        
         Args:
             classifier: The machine learning classifier to optimize.
             data_x (ndarray): 2D array of size (n x m), where n is the
@@ -461,7 +482,7 @@ class Classifier:
             X_train, X_test, y_train, y_test = train_test_split(data, self.data_y, test_size=0.2, random_state=0)
             model0.fit(X_train, y_train)
             y_probas = model0.predict_proba(X_test)
-            skplt.metrics.plot_roc(y_test, y_probas, text_fontsize='large', title='ROC Curve', cmap= 'cividis', plot_macro=False, plot_micro=False)
+            skplt.metrics.plot_roc(y_test, y_probas, text_fontsize='large', title='ROC Curve', cmap='cividis', plot_macro=False, plot_micro=False)
             plt.show()
             return
 
@@ -525,13 +546,20 @@ class Classifier:
         plt.ylabel('True Positive Rate', size=14)
         plt.xlabel('False Positive Rate', size=14)
         plt.title(label=title,fontsize=18)
-        plt.show()
+        if savefig:
+            plt.savefig('Ensemble_ROC_Curve.png', bbox_inches='tight', dpi=300)
+            plt.clf()
+        else:
+            plt.show()
 
-    def plot_hyper_opt(xlog=True, ylog=False):
+    def plot_hyper_opt(self, xlim=None, ylim=None, xlog=True, ylog=False, 
+        savefig=False):
         """
         Plots the hyperparameter optimization history.
     
         Args:
+            xlim: Limits for the x-axis. Ex) xlim = (0, 1000)
+            ylim: Limits for the y-axis. Ex) ylim = (0.9, 0.94)
             xlog (boolean): If True the x-axis will be log-scaled.
                 Defaults to True.
             ylog (boolean): If True the y-axis will be log-scaled.
@@ -541,7 +569,7 @@ class Classifier:
             AxesImage
         """
 
-        fig = optuna.visualization.matplotlib.plot_optimization_history(self.optimization_results)
+        fig = plot_optimization_history(self.optimization_results)
         if xlog:
             plt.xscale('log')
         if ylog:
@@ -549,15 +577,22 @@ class Classifier:
         plt.xlabel('Trial #', size=16)
         plt.ylabel('10-fold CV Accuracy', size=16)
         plt.title(('Hyperparameter Optimization History'), size=18)
-        plt.xlim((1,1e4))
-        plt.ylim((0.9, 0.935))
         plt.xticks(fontsize=14)
         plt.yticks(fontsize=14)
         plt.grid(True, color='k', alpha=0.35, linewidth=1.5, linestyle='--')
-        plt.legend(prop={'size': 16})
-        plt.show()
+        plt.legend(prop={'size': 16}, loc=2)
+        if xlim is not None:
+            plt.xlim(xlim)
+        if ylim is not None:
+            plt.ylim(ylim)
+        plt.tight_layout()
+        if savefig:
+            plt.savefig('Ensemble_Hyperparameter_Optimization.png', bbox_inches='tight', dpi=300)
+            plt.clf()
+        else:
+            plt.show()
 
-    def plot_feature_opt(feats='all'):
+    def plot_feature_opt(self, feats='all'):
         """
         Returns whisker plot displaying the z-score distribution of each feature
         across all trials.
@@ -609,17 +644,18 @@ def evaluate_model(classifier, data_x, data_y, normalize=True, k_fold=10):
 
     return predicted_targets, actual_targets
 
-def generate_matrix(predicted_labels_list, actual_targets, normalize=True, classes=["DIFFUSE","OTHER"], title='Confusion matrix'):
+def generate_matrix(predicted_labels_list, actual_targets, classes, normalize=True, 
+    title='Confusion Matrix', savefig=False):
     """
     Generates the confusion matrix using the output from the evaluate_model() function.
 
     Args:
         predicted_labels_list: 1D array containing the predicted class labels.
         actual_targets: 1D array containing the actual class labels.
+        classes (list): A list containing the label of the two training bags. This
+            will be used to set the axis. Ex) classes = ['DIFFUSE', 'OTHER']
         normalize (bool, optional): If True the matrix accuracy will be normalized
             and displayed as a percentage accuracy. Defaults to True.
-        classes (list): A list containing the label of the two training bags. This
-            will be used to set the axis. Defaults to a list containing 'DIFFUSE' & 'OTHER'. 
         title (str, optional): The title of the output plot. 
 
     Returns:
@@ -634,7 +670,11 @@ def generate_matrix(predicted_labels_list, actual_targets, normalize=True, class
         generate_plot(conf_matrix, classes=classes, normalize=normalize, title=title)
     elif normalize == False:
         generate_plot(conf_matrix, classes=classes, normalize=normalize, title=title)
-    plt.show()
+    if savefig:
+        plt.savefig('Ensemble_Confusion_Matrix.png', bbox_inches='tight', dpi=300)
+        plt.clf()
+    else:
+        plt.show()
 
 def generate_plot(conf_matrix, classes, normalize=False, title='Confusion Matrix'):
     """
@@ -662,6 +702,8 @@ def generate_plot(conf_matrix, classes, normalize=False, title='Confusion Matrix
     tick_marks = np.arange(len(classes))
     plt.xticks(tick_marks, classes, rotation=45, fontsize=14)
     plt.yticks(tick_marks, classes, fontsize=14)
+    #plt.xticks(tick_marks, ['DIFFUSE','OTHER'], rotation=45, fontsize=14)
+    #plt.yticks(tick_marks, ['DIFFUSE','OTHER'], fontsize=14)
 
     fmt = '.2f' if normalize is True else 'd'
     thresh = conf_matrix.max() / 2.
@@ -670,9 +712,9 @@ def generate_plot(conf_matrix, classes, normalize=False, title='Confusion Matrix
         plt.text(j, i, format(conf_matrix[i, j], fmt), fontsize=14, horizontalalignment="center",
                  color="white" if conf_matrix[i, j] > thresh else "black")
 
-    plt.tight_layout()
     plt.ylabel('True label',fontsize=18)
     plt.xlabel('Predicted label',fontsize=18)
+    plt.tight_layout()
 
     return conf_matrix
 
